import sys, os
testdir = os.path.dirname(__file__)
srcdir = '../fosvis'
sys.path.insert(0, os.path.abspath(os.path.join(testdir, srcdir)))

import fosvis
import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import filecmp
import pandas as pd
from pandas.testing import assert_frame_equal
import difflib


"""
To run all: python3 -m unittest
To run a specific test: python3 -m unittest test.test_fosvis
"""

# Done
class test_setup_project(unittest.TestCase):

    def test_setup_project_directory_already_exists(self):
        """
        Tests that when trying to setup a project with the same directory that
        already exists it will throw a SystemExit error.
        """
        test_dir_path = 'test/data_for_testing/testing_temporary_files/setup_project_test_dir'
        os.mkdir(test_dir_path)
        with self.assertRaises(SystemExit):
            fosvis.setup_project_dirs(test_dir_path, 'blastn')
        os.rmdir(test_dir_path)


class test_create_circos_data(unittest.TestCase):

    def test_(self):
        pass


class test_write_paramaters_log(unittest.TestCase):

    def test_write_paramaters_log(self):
        pass


if __name__ == "__main__":
    unittest.main()



"""
Functions that work with seqeunces.
Realtes to karyotype and GC data
"""

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.SeqUtils import GC
from io import StringIO
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import pandas as pd
import itertools
import subprocess
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import statistics
import seaborn as sns
from tqdm import tqdm

def remove_too_small_contigs(contigs, min_contig_length):
    """
    Loop thorugh a fasta file and return a list of BioPython SeqRecord objects that repersent all the seqeunces
    in the fasta file with length >= min_contig_length

    Args:
        contigs (str): A file path to a .fasta file containing contigs
        min_contig_length (int): Contigs smaller than this size won't be added to the returned list

    Returns:
        A list of BioPython SeqRecord objects that repersent only the seqeunces in the contigs file with
        size >= min_contig_length
    """

    print("Removing too small contigs...")
    num_contigs_orginally = 0
    correct_size_contigs = []
    for record in SeqIO.parse(contigs, "fasta"):
        if len(record) >= min_contig_length:
            correct_size_contigs.append(record)
        num_contigs_orginally += 1

    if len(correct_size_contigs) == 0:
        sys.exit("No contigs with length >= min_contig_length were found.")

    print(str(len(correct_size_contigs)) + '/' + str(num_contigs_orginally) + ' contigs were of correct size')

    return correct_size_contigs


def write_correct_len_contigs(project_directory, fosmid_size_contigs):
    """
    Writes all the fosmid_size_contigs to one file and returns that file path and
    writes each seqeunce in fosmid_size_contigs to a file individually

    Args:
        project_directory (str): A file path to the project_directory
        fosmid_size_contigs (list<Biopython.SeqRecord>): A list of BioPython SeqRecord objects that repersent only the seqeunces in the contigs file with
            size >= min_contig_length

    Return:
        correct_size_contigs_path (str): A file path to the file containing all the fosmid_size_contigs together

    """
    # Write all correct size contigs to their own fasta file
    for seq in fosmid_size_contigs:
        seq_path = project_directory + "/intermediate_outputs/correct_length_contigs/" + seq.id + ".fasta"
        SeqIO.write(seq, seq_path, "fasta")

    # writes all the correct_size_contigs to one file
    correct_size_contigs_path = project_directory + "/intermediate_outputs/correct_length_contigs/all_correct_size_contigs.fasta"
    all_seqs = []
    for seq in fosmid_size_contigs:
        all_seqs.append(seq)
    SeqIO.write(all_seqs, correct_size_contigs_path, "fasta")

    return correct_size_contigs_path


def get_karyotype_data(fosmid_size_contigs):
    """
    Creates the karyotype data for the sequences in fosmid_size_contigs in the form of a dataframe.

    Args:
        fosmid_size_contigs (list<SeqRecord>): A list of BioPython SeqRecord objects that the karyotype is to be made for

    Returns:
        karyotype_df (pandas.dataframe): A dataframe containing the karyotype data in the df:
            ['chr_prefix', '-prefix', 'variable_name', 'diagram_label', 'start', 'end', 'color']

    Notes:
    -> chr_prefix col is just 'chr'
    -> -prefix col is just '-'
    -> variable_name is the contig
    -> Diagram label is a number from 1 - # of contigs
    -> start is 1
    -> end is (length + 1) because of circos indexing (i.e. bp 1 is repersented by the range 1 - 2)
    """

    print("Creating karyotype data...")
    karyotype_df = pd.DataFrame(index=[], columns=['chr_prefix', '-prefix', 'variable_name', 'diagram_label', 'start', 'end', 'color'])

    counter = 1
    for seq in fosmid_size_contigs:
        karyotype_row = {'chr_prefix' : 'chr', '-prefix' : '-', 'variable_name' : seq.id, 'diagram_label' : str(counter), 'start' : 1, 'end' : len(seq) + 1, 'color' : 'rgb(120,120,120,0.4)'}
        karyotype_df = karyotype_df.append(karyotype_row, ignore_index=True)
        counter += 1

    return karyotype_df


def write_karyotype_data(project_directory, karyotype_df):
    """
    Writes a df containing karyotype data to a txt file in the project_directory

    Args:
        project_directory (str): A file path to the project_directory
        karyotype_df (pandas.DataFrame): A dataframe contaiing karyotype data

    Returns:
        karyotype_file (str): The file path to where the df was written

    """
    karyotype_file = project_directory + '/circos_diagram_input_data/karyotype.txt'
    karyotype_df.to_csv(karyotype_file, sep=' ', index=False, header=False)
    return karyotype_file


def gc_interval(seqs_file, interval_length):
    """
    Returns a dataframe with the GC content of a seqeunce in interval lengths

    Args:
        seqs_file (str): A file with contigs to determine GC content for (can contain multiple seqs)
        interval_length (int): The length of the interval to determine the GC-content in

    Returns:
        (pandas.core.frame.DataFrame): A dataframe with GC-content over set intervals of a seqeunce in the form:
            ['contig', 'interval_start', 'interval_end', 'gc_content']

    Notes:
    - Bipython seq[start:end] is start (0-based) to end (0-based, non-inlcusive)
    - Due to circos range indexing the GC interval for bp pos x - y would be x - (y + 1)
        - The + 1 is so it covers full range of last bp
        - e.g. gc content for bp's 5 - 10 would be ['contig', 5, 11, 'gc_content']
    """
    print("Creating GC content data...")
    gc_content_df = pd.DataFrame(columns = ['contig', 'interval_start', 'interval_end', 'gc_content'], index = [])

    num_seqs = len([1 for line in open(seqs_file) if line.startswith(">")])
    with tqdm(total=num_seqs) as pbar:
        for sequence in SeqIO.parse(seqs_file, 'fasta'):

            current_interval_start = 0

            while ((current_interval_start + interval_length) < (len(sequence.seq))):
                seq_subset = sequence.seq[current_interval_start:current_interval_start+interval_length]
                gc_content = GC(seq_subset)
                gc_new_row = {'contig' : sequence.id,
                              'interval_start': current_interval_start + 1,
                              'interval_end': (current_interval_start + (interval_length + 1)), 'gc_content': gc_content} # good
                gc_content_df = gc_content_df.append(gc_new_row, ignore_index=True)

                current_interval_start += (interval_length)

            else:
                seq_subset = sequence.seq[current_interval_start:]
                gc_content = GC(seq_subset)
                gc_new_row = {'contig' : sequence.id,
                              'interval_start': current_interval_start + 1,
                              'interval_end': len(sequence.seq) + 1, 'gc_content': gc_content}
                gc_content_df = gc_content_df.append(gc_new_row, ignore_index=True)

            pbar.update(1)


    return gc_content_df



from fosvis import orf

import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import filecmp
import os
import pandas as pd
from pandas.testing import assert_frame_equal
import difflib

# Done
class test_run_prodigal(unittest.TestCase):
    """
    Checks that prodigal creates both output files
    Does not check that contents of the file
    """

    def test_run_prodigal(self):
        sample_seq_path = './test/data_for_testing/test_orf_data/test_orf_sample_seq.fasta'
        output_path = './test/data_for_testing/test_orf_data/temp_outputs/orf_test_output.txt'
        prtn_output_path = './test/data_for_testing/test_orf_data/temp_outputs/orf_test_prtn_output.faa'
        orf.run_prodigal(sample_seq_path, output_path, prtn_output_path)

        # Check both outputs exist and are non-empty
        self.assertTrue(os.path.exists(output_path) and os.path.getsize(output_path) > 0)
        self.assertTrue(os.path.exists(prtn_output_path) and os.path.getsize(prtn_output_path) > 0)

        # remove the files
        os.remove(output_path)
        os.remove(prtn_output_path)


# Done
class test_parse_prodigal_sco_output(unittest.TestCase):

    def test_parse_prodigal_sco_output(self):
        # Parse a sample file
        prodigal_sample_output_path = './test/data_for_testing/test_orf_data/test_parse_prodigal_sco_output_input.txt'
        output = orf.parse_prodigal_sco_output(prodigal_sample_output_path)
        forward_strand_genes_actual_df = output[0]
        reverse_strand_genes_actual_df = output[1]

        #  Create the expected output
        forward_strand_genes_expected_df = pd.DataFrame({'contig' : ['contig_1', 'contig_1', 'contig_1', 'contig_1', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2'],
                                                         'orf_start' : [1,759,1688,2361,18443,19870,20100,22725,23899],
                                                         'orf_end' : [547,1686,2345,3192,19451,20077,20574,23064,25633]})


        reverse_strand_genes_expected_df = pd.DataFrame({'contig' : ['contig_1', 'contig_1', 'contig_1', 'contig_1', 'contig_1', 'contig_1', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2', 'contig_2'],
                                                         'orf_start' : [3228,4094,5473,7205,7623,8844,95,1768,2273,11474,12047,13858,14829,15125,17691,17807],
                                                         'orf_end' : [4095,5477,7108,7532,8835,9141,1145,2245,2999,11777,12371,14521,15015,16082,17811,18038]})


        # Check that actual and expected have same shape
        self.assertEqual(forward_strand_genes_expected_df.shape, forward_strand_genes_actual_df.shape)
        self.assertEqual(reverse_strand_genes_expected_df.shape, reverse_strand_genes_actual_df.shape)

        # Check contents is equal
        self.assertEqual(assert_frame_equal(forward_strand_genes_expected_df, forward_strand_genes_actual_df), None)
        self.assertEqual(assert_frame_equal(reverse_strand_genes_expected_df, reverse_strand_genes_actual_df), None)




if __name__ == "__main__":
    unittest.main()



"""
All functions related to protein_domain data and HMMER
"""

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.SeqUtils import GC
from io import StringIO
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import pandas as pd
import itertools
import subprocess
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import statistics
import seaborn as sns


def run_hmmscan(orfs, hmm_db, output_dir, E_value, prefix):
    """
    Runs hmmscan on open reading frames (ORFs) using a specified (pressed) HMM database

    Args:
        orfs (str): File path to .faa file containing all the ORFs
        hmm_db (str): File path to a pressed HMM database
        output_dir (str): Directory to to put all output files in (should already exist)
        E_value (int): The E value used when running hmmscan
        prefix (str): Prefix for all the output files (e.g. the sample name)

    Returns:
        None
    """

    out = output_dir + '/' + prefix + '_out.txt'
    tblout = output_dir + '/' + prefix + '_tblout.txt'
    domtblout = output_dir + '/' + prefix + '_domtblout.txt'
    hmmscan_stdout_and_stderr_log_path = output_dir + '/hmmscan_stdout_and_stderr_log.txt'

    hmmscan_command = 'hmmscan -o ' + out + ' --tblout ' + tblout + ' --domtblout ' + domtblout + ' -E ' + str(E_value) + ' ' + hmm_db + ' ' + orfs

    with open(hmmscan_stdout_and_stderr_log_path, "wb") as stdout:
        subprocess.call([str(hmmscan_command)], stdout=stdout, stderr=subprocess.STDOUT, shell=True)


def parse_hmmscan_domtblout(domtblout):
    """
    Parse a hmmscan domtbl output file and return a pandas dataframe with the output.
    Designed to be used for a hmmscan output file that was for hmmscan run on a prodigal output.

    Args:
        domtblout (str): File path to a hmmscan domtbl output file to be parsed

    Returns:
        (pandas df): Contains the parsed domtblout information, in the form:
            ['target', 'orf_id', 'orf_start', 'orf_stop', 'target_description']
    """

    with open(domtblout, 'r') as f:

        hmmscan_domains = pd.DataFrame(index=[], columns=['target', 'orf_id', 'orf_start', 'orf_stop', 'target_description'])

        for line in f:

            if (line[0] == '#'):
                continue

            target = line.split()[0]
            orf_id = line.split()[3]
            orf_start = line.split()[17]
            orf_stop = line.split()[18]
            target_description = ' '.join(line.split()[22:])

            new_target_entry = {'target' : target, 'orf_id': orf_id, 'orf_start' : orf_start, 'orf_stop' : orf_stop, 'target_description' : target_description}
            hmmscan_domains = hmmscan_domains.append(new_target_entry, ignore_index=True)

    return hmmscan_domains


def convert_hmmscan_domains_from_protein_to_nuc_pos(hmmscan_df, orf_file):
    """
    Takes in a hmmscan dataframe and determines the nucleotide start/stop positions
    of the idneitfied protein domain in terms of its protein start/stop pos given by the
    hmmscan output.

    Also colors the protein domains the same if they are the same domain.

    Args:
        hmmscan_df (pandas.dataframe): A dataframe from a parsed hmmscan domtbl file
            Needs to be in the following format: ['target', 'orf_id', 'orf_start', 'orf_stop', 'target_description']
        orf_file (str): File path to prodigal orf protein seqeunce file
            Headers need to be in the following format (format from prodigal):
            <contig>_<orf number identifed> # <bp start> # <bp stop> # <strand> # etc.

    Returns:
        (pandas.dataframe): A dataframe in the following fomrat ['band', 'contig', 'target_var', 'target_label', 'start', 'stop', 'color']
    """

    # Create dictionary of orf_id and [start, stop, strand]
    orf_start_stop_dict = {}
    for orf in SeqIO.parse(orf_file, 'fasta'):
        orf_start = orf.description.split()[2]
        orf_stop = orf.description.split()[4]
        forward_or_reverse = orf.description.split()[6] # 1 or -1
        orf_start_stop_dict [orf.id] = [orf_start, orf_stop, forward_or_reverse]

    band_data_df = pd.DataFrame(index=[], columns=['band', 'contig', 'target_var', 'target_label', 'start', 'stop', 'color'])

    # Make Dictionary with keys of every unique domain and values as a random rgb color
    # loop thorugh every row of dataframe and make list of all the domains and turn that into a set
    unique_domains = {}
    for index, row in hmmscan_df.iterrows():
        random_color = random_rgb_color()
        unique_domains[row['target']] = random_color


    for index, row in hmmscan_df.iterrows():

        # Get the unique color for that specific target (all bands with that target will be same color)
        band_color = unique_domains[row['target']]

        contig = row['orf_id'].rsplit('_', 1)[0]

        orf_bp_start = orf_start_stop_dict[row['orf_id']][0]
        orf_bp_stop = orf_start_stop_dict[row['orf_id']][1]
        orf_forward_or_reverse = orf_start_stop_dict[row['orf_id']][2]

        domain_bp_start = None
        domain_bp_stop = None

        # ORF on forward strand
        if int(orf_forward_or_reverse) == 1:
            domain_bp_start = int(orf_bp_start) + 3 * (int(row['orf_start']) - 1)
            domain_bp_stop = int(orf_bp_start) + 3 * int(row['orf_stop']) - 1
        # ORF on reverse strand
        elif int(orf_forward_or_reverse) == -1:
            domain_bp_stop = int(orf_bp_stop) - 3 * (int(row['orf_start']) - 1)
            domain_bp_start = int(orf_bp_stop) - 3 * int(row['orf_stop']) + 1

        target_description_spaces_to_underscore = row['target_description'].replace(' ', '_')

        new_band = {'band' : 'band', 'contig' : contig, 'target_var' : row['target'], 'target_label' : target_description_spaces_to_underscore, 'start' : str(domain_bp_start), 'stop' : str(domain_bp_stop), 'color' : band_color}
        band_data_df = band_data_df.append(new_band, ignore_index=True)

    return band_data_df


def get_prtn_domain_band_data(orfs, hmm_db, output_dir, E_value, prefix):
    """
    Calls functions to run hmmscan, parse the output, convert the start/stop positions
    of the domains from positions on the ORFs to positions on the original contig and
    creates a dataframe with the data in the correct format for circos band data.
    This function is meant to be used with ORFs with headers created by the prodigal tool.

    Args:
        orfs (str): File path to .faa file containing all the ORFs (should be created by prodigal and in the prodigal format)
        hmm_db (str): File path to a pressed HMM database
        output_dir (str): Directory to to put all hmmscan output files in (should already exist)
        E_value (str): The E value used when running hmmscan
        prefix (str): Prefix for all the output files (e.g. the sample name)

    Returns:
        (pandas.dataframe): A dataframe with the band data in the form:
            ['band', 'contig', 'target_var', 'target_label', 'start', 'stop', 'color']
    """
    print("Creating protein domain band data - this may take a while depending on amount/size of fosmids...")
    run_hmmscan(orfs, hmm_db, output_dir, E_value, prefix)
    parsed_hmm_df = parse_hmmscan_domtblout(output_dir + "/" + prefix + "_domtblout.txt")
    return convert_hmmscan_domains_from_protein_to_nuc_pos(parsed_hmm_df, orfs)


def random_rgb_color():
    """
    Generate a random RGB Color.

    Returns:
        (str): A string in the fomrat 'rgb(x,y,z)'
    """

    color = list(np.random.choice(range(256), size=3))
    return 'rgb(' + str(color[0]) + "," + str(color[1]) + ',' + str(color[2]) + ')'



from fosvis import diagram_and_legend

import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import filecmp
import os
import pandas as pd
from pandas.testing import assert_frame_equal
import difflib


class test_create_circos_conf_file(unittest.TestCase):

    def test_create_circos_conf_file_no_gc(self):
        no_gc_actual_ouput_path = 'test/data_for_testing/test_diagram_and_legend/test_create_circos_conf_file/circos_actual_no_gc.conf'
        diagram_and_legend.create_circos_conf_file(no_gc_actual_ouput_path, 'karyotype_path.txt', 'links_path.txt', 'ORF_path.txt', 'ORF_reverse_path.txt')
        expected_output_file_no_gc = 'test/data_for_testing/test_diagram_and_legend/test_create_circos_conf_file/circos_expected_no_gc.conf'
        self.assertTrue(filecmp.cmp(expected_output_file_no_gc, no_gc_actual_ouput_path, shallow=False))
        os.remove(no_gc_actual_ouput_path)


    def test_create_circos_conf_file_with_gc(self):
        with_gc_actual_ouput_path = 'test/data_for_testing/test_diagram_and_legend/test_create_circos_conf_file/circos_actual_with_gc.conf'
        diagram_and_legend.create_circos_conf_file(with_gc_actual_ouput_path, 'karyotype_path.txt', 'links_path.txt', 'ORF_path.txt', 'ORF_reverse_path.txt', gc_data='gc_content.txt')
        expected_output_file_with_gc = 'test/data_for_testing/test_diagram_and_legend/test_create_circos_conf_file/circos_expected_with_gc.conf'
        self.assertTrue(filecmp.cmp(expected_output_file_with_gc, with_gc_actual_ouput_path, shallow=False))
        os.remove(with_gc_actual_ouput_path)



class test_make_diagram(unittest.TestCase):

    def test_(self):
        pass


class test_prtn_domain_legend(unittest.TestCase):

    def test_(self):
        pass


class test_draw_prtn_domain_legend(unittest.TestCase):

    def test_(self):
        pass


class test_karyotype_legend(unittest.TestCase):

    def test_(self):
        pass


if __name__ == "__main__":
    unittest.main()



from fosvis import seq_util

import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import filecmp
import os
import pandas as pd
from pandas.testing import assert_frame_equal
import difflib


# Done
class test_remove_too_small_contigs(unittest.TestCase):

    def test_remove_too_small_contigs(self):
        input_file = 'test/data_for_testing/Fosmid_Size_Selection_Tests/test_remove_too_small_contigs_input.fasta'
        result = seq_util.remove_too_small_contigs(input_file, 100)

        seq3_len_100 = SeqRecord(
            Seq("GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG"),
            id="seq3_len_100",
            name="seq3_len_100",
            description="seq3_len_100")
        seq4_len_101 = SeqRecord(
            Seq("GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG"),
            id="seq4_len_101",
            name="seq4_len_101",
            description="seq4_len_101")
        seq5_len_200 = SeqRecord(
            Seq("GGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGGG"),
            id="seq5_len_200",
            name="seq5_len_200",
            description="seq5_len_200")
        expected_result = [seq3_len_100, seq4_len_101, seq5_len_200]

        self.assertEqual(len(result), len(expected_result))
        self.assertTrue(result[0].seq == expected_result[0].seq)
        self.assertTrue(result[1].seq == expected_result[1].seq)
        self.assertTrue(result[2].seq == expected_result[2].seq)
        self.assertTrue(result[0].id == expected_result[0].id)
        self.assertTrue(result[1].id == expected_result[1].id)
        self.assertTrue(result[2].id == expected_result[2].id)

# Done
# class test_write_seqs_to_file(unittest.TestCase):
#
#     def test_write_seqs_to_file(self):
#         file_to_write = 'test/data_for_testing/Fosmid_Size_Selection_Tests/write_seqs_to_file_actual_output.fasta'
#         seq1 = SeqRecord(
#             Seq("AAGGTTCC"),
#             id="seq1",
#             name="seq1",
#             description="seq1")
#         seq2 = SeqRecord(
#             Seq("GGAACCTT"),
#             id="seq2",
#             name="seq2",
#             description="seq2")
#         seqs_to_write = [seq1,seq2]
# 
#         seq_util.write_seqs_to_file(seqs_to_write, file_to_write)
#
#         expected_output_file = 'test/data_for_testing/Fosmid_Size_Selection_Tests/test_write_seqs_to_file_expected_output.fasta'
#
#         self.assertTrue(filecmp.cmp(expected_output_file, file_to_write, shallow=False))
#
#         os.remove(file_to_write)

# Done
class test_get_karyotype_data(unittest.TestCase):

    def test_get_karyotype_data(self):
        seq1_len_8 = SeqRecord(
            Seq("AAGGTTCC"),
            id="seq1_len_8",
            name="seq1_len_8",
            description="seq1_len_8")
        seq2_len_18 = SeqRecord(
            Seq("GGAACCTTGGAACCTT"),
            id="seq1_len_8",
            name="seq1_len_8",
            description="seq1_len_8")
        seqs = [seq1_len_8,seq2_len_18]
        result = seq_util.get_karyotype_data(seqs)

        expected_result_data = {'chr_prefix':['chr', 'chr'], '-prefix':['-', '-'], 'variable_name':['seq1_len_8', 'seq1_len_8'], 'diagram_label':['1', '2'], 'start':[1, 1], 'end':[9, 17], 'color':['rgb(120,120,120,0.4)', 'rgb(120,120,120,0.4)']}
        expected_result_df = pd.DataFrame(expected_result_data)

        self.assertEqual(assert_frame_equal(result, expected_result_df, check_dtype=False), None)

# Done
class test_gc_interval(unittest.TestCase):

    def test_gc_interval(self):

        interval_10_input_file = 'test/data_for_testing/gc_interval_tests/gc_content_test_interval_10_input.fasta'
        interval_3_input_file = 'test/data_for_testing/gc_interval_tests/gc_content_test_interval_3_input.fasta'

        interval_10_result = seq_util.gc_interval(interval_10_input_file, 10)
        interval_3_result = seq_util.gc_interval(interval_3_input_file, 3)

        expected_interval_10_result = {'contig':['interval_10_mix', 'interval_10_mix', 'interval_10_mix', 'interval_10_mix',
                                                 'interval_10_all_GC', 'interval_10_all_GC', 'interval_10_not_divis_by_10'],
                                       'interval_start':[1, 11, 21, 31, 1, 11, 1], 'interval_end':[11,21,31,41, 11, 21, 5], 'gc_content':[100,50,50,0,100,100,50]}
        expected_interval_10_result_df = pd.DataFrame(expected_interval_10_result)
        expected_interval_3_result = {'contig':['interval_3_mix', 'interval_3_mix', 'interval_3_mix', 'interval_3_mix',
                                                 'interval_3_not_divis_by_3', 'interval_3_not_divis_by_3', 'interval_3_not_divis_by_3', 'interval_3_not_divis_by_3'],
                                       'interval_start':[1,4,7,10, 1, 4, 7, 10],
                                       'interval_end':[4, 7, 10, 13, 4, 7, 10, 11],
                                       'gc_content':[((2/3)*100), 0, 100, (1/3)*100, 100,100,0,100]}

        expected_interval_3_result_df = pd.DataFrame(expected_interval_3_result)

        self.assertEqual(assert_frame_equal(interval_10_result, expected_interval_10_result_df, check_dtype=False), None)
        self.assertEqual(assert_frame_equal(interval_3_result, expected_interval_3_result_df, check_dtype=False), None)


if __name__ == "__main__":
    unittest.main()



from fosvis import seq_util
from fosvis import links
from fosvis import protein_domains
from fosvis import orf
from fosvis import diagram_and_legend

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.SeqUtils import GC
from io import StringIO
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import pandas as pd
import itertools
import subprocess
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import statistics
import seaborn as sns
from shutil import which

"""
How to use: Script is used through the create_circos_data() and make_diagram() methods.  See documentation pdf for more detail.
"""

############################### create_circos_data #############################

def create_circos_data(contigs, output_dir, project_title, orfs=True, gc=True, custom_histogram=False, custom_histogram_file='', hmm_db='', e_value=0.01, min_contig_length=10000, min_blast_similarity_percentage=90, min_blast_similarity_length=300, link_transperacny='0.60', percent_link_overlap_tolerance=50, include_domains=False, gc_interval_len=100, blast_type='blastn', keep_blast_data=False):
    """
    Create the following input files for a circos diagram:
        - ORF.txt
        - ORF_reverse.txt
        - links.txt
        - karyotype.txt
        - circos.conf

    Args:
        contigs (str): File path to a FASTA file containing contigs
        output_dir (str): File path to a directory where the project folder will be created
        project_title (str): Name of the project directory that is created in the ouput_dir (and used for some file prefixes)
        orfs (bool): If True will include orfs layers, if False don't
        gc (bool): If True include GC layer, if False don't
        custom_histogram (bool): If True, inlcude custom histogram layer, if False don't
        custom_histogram_file (str): File path to a custom histogram data txt file (see documentation for more details)
        hmm_db (str): File path to a pressed hmm database
        e_value (int): The e-value used for the hmmscan search
        min_contig_length (int): The minimum size of a contig to be used in the diagram (idea is that each contig should represent one fosmid)
        min_blast_similarity_percentage (int): Min nucleotide percent similarity (blast percent identity) for link data
        min_blast_similarity_length (int): Min nucleotide similarity length (blast align length) for link data
        link_transperacny (str): The transparency of the links in the diagram in range 0 - 1 (0 is transparent)
        percent_link_overlap_tolerance (int): The amount of overlap percentage required to classify two links as similar enough to color them with the same color
        inlcude_domains (boolean): If True will create protein domain band data, if false will not
        gc_interval_len (int): The interval over which gc content is calculated for the histogram
        blast_type (str): The type of blast to use to create the links (can be 'blastn' OR 'tblastx')
        keep_blast_data (bool): If True will keep the raw blast data, won't if false

    Returns:
        None
    """
    # check if all required programs are in PATH
    if which('circos') is None:
        print("Could not find circos, please install and append to PATH")
        sys.exit()

    if which('prodigal') is None:
        print("Could not find prodigal, please install and append to PATH")
        sys.exit()

    if blast_type == 'blastn':
        if which('blastn') is None:
            print("Could not find blastn, please install and append to PATH")
            sys.exit()

    if blast_type == 'tblastx':
        if which('bl2seq') is None:
            print("Could not find tblastx (called bl2seq in PATH), please install and append to PATH")
            sys.exit()

    if which('hmmscan') is None:
        print("Could not find hmmscan, please install and append to PATH")
        sys.exit()

    project_directory = output_dir + '/' + project_title

    setup_project_dirs(project_directory, blast_type)

    write_paramters_log(project_directory, contigs, output_dir, project_title, hmm_db, e_value, min_contig_length, min_blast_similarity_percentage, min_blast_similarity_length, link_transperacny, percent_link_overlap_tolerance, include_domains, gc, gc_interval_len, blast_type, keep_blast_data)

    # Get correct len contigs
    fosmid_size_contigs = seq_util.remove_too_small_contigs(contigs, min_contig_length)
    correct_size_contigs_path = seq_util.write_correct_len_contigs(project_directory, fosmid_size_contigs)

    # karyotype data
    karyotype_df = seq_util.get_karyotype_data(fosmid_size_contigs)
    karyotype_file = seq_util.write_karyotype_data(project_directory, karyotype_df)

    # Get list of all the correct len contigs
    list_of_correct_len_contigs_paths = os.listdir(project_directory + '/intermediate_outputs/correct_length_contigs')
    list_of_correct_len_contigs_paths.remove('all_correct_size_contigs.fasta')
    list_of_correct_len_contigs_paths = [project_directory + '/intermediate_outputs/correct_length_contigs/' + s for s in list_of_correct_len_contigs_paths]

    # links data
    links_df = links.get_link_data(list_of_correct_len_contigs_paths, blast_type, project_directory, min_blast_similarity_percentage, min_blast_similarity_length, link_transperacny, percent_link_overlap_tolerance, keep_blast_data)
    if links_df.empty:
        print('No links were found, your diagram will still be created but no links will be displayed between your sequences')
    links_df.to_csv(project_directory + '/circos_diagram_input_data/links.txt', sep=' ', index=False, header=False)

    # orf data
    if orfs:
        prodigal_prtn_seq_output = orf.create_orf_data(project_directory, correct_size_contigs_path)

    # Protein domain data
    if include_domains:
        prtn_domain_df = protein_domains.get_prtn_domain_band_data(prodigal_prtn_seq_output, hmm_db, project_directory + '/intermediate_outputs/hmmscan', e_value, project_title)
        prtn_domain_df.to_csv(karyotype_file, mode='a', sep=' ', index=False, header=False)

    # GC data
    if gc:
        gc_content_file = project_directory + '/circos_diagram_input_data/gc_content.txt'
        gc_content_df = seq_util.gc_interval(correct_size_contigs_path, gc_interval_len)
        gc_content_df.to_csv(gc_content_file, sep=' ', index=False, header=False)

    # Create circos conf file
    diagram_and_legend.create_circos_conf_file(project_directory + '/circos_diagram_input_data/circos.conf', project_directory + '/circos_diagram_input_data/karyotype.txt', project_directory + '/circos_diagram_input_data/links.txt', orfs, gc, custom_histogram, orf_forward_file=project_directory + '/circos_diagram_input_data/ORF.txt', orf_reverse_file=project_directory + '/circos_diagram_input_data/ORF_reverse.txt', gc_file=project_directory + '/circos_diagram_input_data/gc_content.txt', custom_histogram_file=custom_histogram_file)
    # if gc:
    #     diagram_and_legend.create_circos_conf_file(project_directory + '/circos_diagram_input_data/circos.conf', project_directory + '/circos_diagram_input_data/karyotype.txt', project_directory + '/circos_diagram_input_data/links.txt', project_directory + '/circos_diagram_input_data/ORF.txt', project_directory + '/circos_diagram_input_data/ORF_reverse.txt', gc_data=project_directory + '/circos_diagram_input_data/gc_content.txt')
    # else:
    #     diagram_and_legend.create_circos_conf_file(project_directory + '/circos_diagram_input_data/circos.conf', project_directory + '/circos_diagram_input_data/karyotype.txt', project_directory + '/circos_diagram_input_data/links.txt', project_directory + '/circos_diagram_input_data/ORF.txt', project_directory + '/circos_diagram_input_data/ORF_reverse.txt')


def setup_project_dirs(project_directory, blast_type):
    """
    Sets up project directory for the project.

    Args:
        project_directory (str): A file path to a directory (the directory for the project)
            that does not already exist yet.
        blast_type (str): The type of blast to use to create the links (can be 'blastn' OR 'tblastx')


    Returns:
        None
    """
    print("Setting up project...")
    try:
        os.mkdir(project_directory)
    except Exception as e:
        print(e)
        print("Was not able to create project directory.")
        sys.exit()

    try:
        os.mkdir(project_directory + '/circos_diagram_input_data')
        os.mkdir(project_directory + '/intermediate_outputs')
        os.mkdir(project_directory + '/intermediate_outputs/prodigal')
        os.mkdir(project_directory + '/intermediate_outputs/correct_length_contigs')
        os.mkdir(project_directory + '/intermediate_outputs/hmmscan')
        if (blast_type == 'blastn'):
            os.mkdir(project_directory + '/intermediate_outputs/blastn')
        elif (blast_type == 'tblastx'):
            os.mkdir(project_directory + '/intermediate_outputs/tblastx')
    except Exception as e:
        print(e)
        print("Was not able to set up internal project directory structure.")
        sys.exit()


def write_paramters_log(project_directory, contigs, output_dir, project_title, hmm_db, e_value, min_contig_length, min_blast_similarity_percentage, min_blast_similarity_length, link_transperacny, percent_link_overlap_tolerance, include_domains, gc, gc_interval_len, blast_type, keep_blast_data):
    """
    Writes the paramters passed into create_circos_data() into the paramters_log.txt

    Args:
        project_directory (str): A file path to the project_directory
        contigs (str): File path to a FASTA file containing contigs
        output_dir (str): File path to a directory where the project folder will be created
        project_title (str): Name of the project directory that is created in the ouput_dir (and used for some file prefixes)
        hmm_db (str): File path to a pressed hmm database
        e_value (int): The e-value used for the hmmscan search
        min_contig_length (int): The minimum size of a contig to be used in the diagram (idea is that each contig should represent one fosmid)
        min_blast_similarity_percentage (int): Min nucleotide percent similarity (blast percent identity) for link data
        min_blast_similarity_length (int): Min nucleotide similarity length (blast align length) for link data
        link_transperacny (str): The transparency of the links in the diagram in range 0 - 1 (0 is transparent)
        percent_link_overlap_tolerance (int): The amount of overlap percentage required to classify two links as similar enough to color them with the same color
        inlcude_domains (boolean): If True will create protein domain band data, if false will not
        gc (boolean): If True the diagram will include a track for GC content, if False it will not
        gc_interval_len (int): The interval over which gc content is calculated for the histogram
        blast_type (str): The type of blast to use to create the links (can be 'blastn' OR 'tblastx')
        keep_blast_data (bool): If True will keep the raw blast data, won't if false

    Returns:
        None
    """
    paramters_log_file_path = project_directory + '/paramaters_log.txt'
    with open(paramters_log_file_path, "x") as f:
        f.write("contigs: " + contigs + '\n'
                "output_dir: " + output_dir + '\n'
                "project_title: " + project_title + '\n'
                "hmm_db: " + hmm_db + '\n'
                "e_value: " + str(e_value) + '\n'
                "min_contig_length: " + str(min_contig_length) + '\n'
                "min_blast_similarity_percentage: " + str(min_blast_similarity_percentage) + '\n'
                "min_blast_similarity_length: " + str(min_blast_similarity_length) + '\n'
                "link_transperacny: " + link_transperacny + '\n'
                "percent_link_overlap_tolerance: " + str(percent_link_overlap_tolerance) + '\n'
                "include_domains: " + str(include_domains) + '\n'
                "gc: " + str(gc) + '\n'
                "gc_interval_len: " + str(gc_interval_len) + '\n'
                "blast_type: " + blast_type + '\n'
                "keep_blast_data: " + str(keep_blast_data) + '\n')
        f.close()


#################### make_diagram ##############################################

def make_diagram(data_dir, ncol=2):
    """
    Makes call to circos to make the diagram.
    Creates the protein domain and karyotype legends.

    Args:
        data_dir (str): File path to a directory containing the following files:
            -> ORF.txt
            -> ORF_reverse.txt
            -> links.txt
            -> karyotype.txt
            -> circos.conf
        ncol (int): Number of columns in the protein_domain legend

    Returns:
        None

    Notes:
        -> Running with -nonparanoid flag so band overlap errors won't terminate the program
    """
    print("Creating circos diagram...")

    circos_stdout_and_stderr_log_path = os.path.dirname(data_dir) + '/intermediate_outputs/circos_stdout_and_stderr_log.txt'
    circos_command = "circos -conf " + data_dir + "/circos.conf -outputdir " + data_dir + " -outputfile circos_diagram.png -noparanoid"

    with open(circos_stdout_and_stderr_log_path,"wb") as stdout:
        subprocess.call([str(circos_command)], stdout=stdout, stderr=subprocess.STDOUT, shell=True)

    diagram_and_legend.prtn_domain_legend(data_dir + "/karyotype.txt", data_dir + "/protein_domain_legend.png", ncol)
    diagram_and_legend.karyotype_legend(data_dir + "/karyotype.txt", data_dir + "/karyotype_legend.png")



from fosvis import links

import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import filecmp
import os
import pandas as pd
from pandas.testing import assert_frame_equal
import difflib


class test_get_link_data(unittest.TestCase):

    def test_get_link_data_blastn(self):
        fake_proj_dir = './test/data_for_testing/links_testing/sample_proj_dir'
        list_of_correct_len_contigs_paths = os.listdir('./test/data_for_testing/links_testing/sample_proj_dir/intermediate_outputs/correct_length_contigs')
        list_of_correct_len_contigs_paths = [fake_proj_dir + '/intermediate_outputs/correct_length_contigs/' + s for s in list_of_correct_len_contigs_paths]

        actual_df = links.get_link_data(list_of_correct_len_contigs_paths, 'blastn', fake_proj_dir, 80, 100, '0.60', 50, False)

        # check shape is right
        self.assertEqual(actual_df.shape, (48,7))

        # check column is right
        actual_columns = list(actual_df.columns.values)
        expected_columns = ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos', 'color_thickness']
        self.assertEqual(actual_columns, expected_columns)

    def test_get_link_data_tblastx(self):
        fake_proj_dir = './test/data_for_testing/links_testing/sample_proj_dir'
        list_of_correct_len_contigs_paths = os.listdir('./test/data_for_testing/links_testing/sample_proj_dir/intermediate_outputs/correct_length_contigs')
        list_of_correct_len_contigs_paths = [fake_proj_dir + '/intermediate_outputs/correct_length_contigs/' + s for s in list_of_correct_len_contigs_paths]

        actual_df = links.get_link_data(list_of_correct_len_contigs_paths, 'tblastx', fake_proj_dir, 80, 100, '0.60', 50, False)

        # check shape is right
        self.assertEqual(actual_df.shape, (371,7))

        # check column is right
        actual_columns = list(actual_df.columns.values)
        expected_columns = ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos', 'color_thickness']
        self.assertEqual(actual_columns, expected_columns)


class test_get_pair_links(unittest.TestCase):

    def test_get_pair_blastn_links(self):
        fake_proj_dir = './test/data_for_testing/links_testing/sample_proj_dir'
        query_path = './test/data_for_testing/links_testing/test_get_pair_links/E01.fasta'
        subject_path =  './test/data_for_testing/links_testing/test_get_pair_links/F01.fasta'
        actual_df = links.get_pair_blastn_links(query_path, subject_path, fake_proj_dir, 50, 50, False)

        # check shape is right
        self.assertEqual(actual_df.shape, (16,6))

        # check column is right
        actual_columns = list(actual_df.columns.values)
        expected_columns = ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos']
        self.assertEqual(actual_columns, expected_columns)


    def test_get_pair_tblastx_links(self):
        fake_proj_dir = './test/data_for_testing/links_testing/sample_proj_dir'
        query_path = './test/data_for_testing/links_testing/test_get_pair_links/query.fasta'
        subject_path =  './test/data_for_testing/links_testing/test_get_pair_links/subject.fasta'
        actual_df = links.get_pair_tblastx_links(query_path, subject_path, fake_proj_dir, 50, 50, False)

        # check shape is right
        self.assertEqual(actual_df.shape, (15,6))

        # check column is right
        actual_columns = list(actual_df.columns.values)
        expected_columns = ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos']
        self.assertEqual(actual_columns, expected_columns)


class test_color_similar_links(unittest.TestCase):

    def test_color_similar_links_shape_and_columns(self):

        # load sample data - csv that repersents a df with homology_link_data
        sample_homology_links_df = pd.read_csv('./test/data_for_testing/links_testing/homology_links_pre_color_similar_links_df.csv')

        actual_df = links.color_similar_links(sample_homology_links_df, '0.60', 50)

        # check shape is right
        self.assertEqual(actual_df.shape[1], 7)
        self.assertTrue(actual_df.shape[0] > 0)

        # check column is right
        actual_columns = list(actual_df.columns.values)
        expected_columns = ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos', 'color_thickness']
        self.assertEqual(actual_columns, expected_columns)


class test_merge_lists_with_common_elements(unittest.TestCase):

    def test_merge_lists_with_common_elements(self):
        # Most cases done

        # One item
        listoflist = [[1,2,3]]
        result = links.merge_lists_with_common_elements(listoflist)
        expected_result = [{1,2,3}]
        self.assertEqual(result, expected_result)

        # 4 items - none merged
        listoflist = [[1,2], [3,4], [5,6], [7,8]]
        result = links.merge_lists_with_common_elements(listoflist)
        expected_result = [{1,2}, {3,4}, {5,6}, {7,8}]
        self.assertEqual(result, expected_result)

        # 4 items - all merged
        listoflist = [[1,2,3,4], [4,5,6,7,], [7,8,9,10], [10,11]]
        result = links.merge_lists_with_common_elements(listoflist)
        expected_result = [{1,2,3,4,5,6,7,8,9,10,11}]
        self.assertEqual(result, expected_result)

        # Two items being merged
        listoflist = [[1,2,3], [7,3,9]]
        result = links.merge_lists_with_common_elements(listoflist)
        expected_result = [{1,2,3,7,3,9}]
        self.assertEqual(result, expected_result)

        # Two items being merged, one not
        listoflist = [[1,2,3], [7,3,9], [11,22,33]]
        result = links.merge_lists_with_common_elements(listoflist)
        expected_result = [{1,2,3,7,3,9}, {11,22,33}]
        self.assertEqual(result, expected_result)

        # Multiple merges
        listoflist = [[1,2,3], [7,3,9], [11,22,33], [7,77,777]]
        result = links.merge_lists_with_common_elements(listoflist)
        expected_result = [{1,2,3,7,3,9,7,77,777}, {11,22,33}]
        self.assertEqual(result, expected_result)


class test_calc_percentage_overlap(unittest.TestCase):

    def test_calc_percentage_overlap(self):

        # No overlap
        range1 = (0,10)
        range2 = (10,20)
        result = links.calc_percentage_overlap(range1, range2)
        self.assertEqual(result, 0.0)

        # Normal partial overlap
        range1 = (0,10)
        range2 = (5,15)
        result = links.calc_percentage_overlap(range1, range2)
        self.assertEqual(result, 33.33333333333333)

        # # Normal partial overlap - range1 inverted
        range1 = (10,0)
        range2 = (5,15)
        result = links.calc_percentage_overlap(range1, range2)
        self.assertEqual(result, 33.33333333333333)

        # Normal partial overlap - range1 and range2 inverted
        range1 = (10,0)
        range2 = (15,5)
        result = links.calc_percentage_overlap(range1, range2)
        self.assertEqual(result, 33.33333333333333)

        # Normal partial overlap - range2 inverted
        range1 = (0,10)
        range2 = (15,5)
        result = links.calc_percentage_overlap(range1, range2)
        self.assertEqual(result, 33.33333333333333)

        # Complete overlap
        range1 = (0,100)
        range2 = (10,20)
        result = links.calc_percentage_overlap(range1, range2)
        self.assertEqual(result, 10)


class test_create_color_palette(unittest.TestCase):

    def test_create_color_palette(self):

        # 1 color
        result_1 = links.create_color_palette(1)
        expected_result_1 = ['219.29999999999998,94.65599999999999,86.69999999999999']
        self.assertEqual(result_1, expected_result_1)
        self.assertTrue(all(isinstance(n, str) for n in result_1))

        # 3 color
        result_3 = links.create_color_palette(3)
        expected_result_3 = ['219.29999999999998,94.65599999999999,86.69999999999999',
                             '86.69999999999999,219.29999999999998,94.65599999999999',
                             '94.65599999999999,86.69999999999999,219.29999999999998']
        self.assertEqual(result_3, expected_result_3)
        self.assertTrue(all(isinstance(n, str) for n in result_3))

        # 10 color
        result_10 = links.create_color_palette(10)
        self.assertTrue(all(isinstance(n, str) for n in result_10))





if __name__ == "__main__":
    unittest.main()



"""All functions related to creating diagrams and legends"""

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.SeqUtils import GC
from io import StringIO
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import pandas as pd
import itertools
import subprocess
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import statistics
import seaborn as sns



def find_layer_radii(orfs, gc, custom_histogram):
    """
    Creates a dictionary with the proper circos radii for the given set of layer inputs to be used in the conf file

    Args:
        orfs (bool): If True will include orfs layers, if False don't
        gc (bool): If True include GC layer, if False don't
        custom_histogram (bool): If True, inlcude custom histogram layer, if False don't

    Returns:
        (tuple(int, dict)): A tuple with the first value being an int that corresponds to a layer combination
                            (as seen in the notes section below) and the second value is a dictioary that corresponds
                            to the correct circos radius positions for that layer combiantion

    Notes:
        Options:
        0 = karyotype, links
        1 = karyotype, ORFs, links
        2 = karyotype, GC, links
        3 = karyotype, custom_histogram, links
        4 = karyotype, ORFs, GC, links
        5 = karyotype, ORFs, custom_histogram, links
        6 = karyotype, GC, custom_histogram, links
        7 = karyotype, ORFs, GC, custom_histogram, links
    """

    # possible combiantions of layer radii
    layer_radii_combinations = [{'karyotype' : '0.80r', 'links' : '0.96r'},
                                {"karyotype": '0.80r', "ORF": ('0.91r', '0.96r'), "ORF_reverse" : ('0.83r', '0.88r'), "links" : '0.79r'},
                                {"karyotype": '0.80r', "GC": ('0.84r', '0.96r'), "links" : '0.80r'},
                                {"karyotype": '0.80r', "custom_histogram" : ('0.84r', '0.96r'), "links" : '0.80r'},
                                {"karyotype": '0.80r', "ORF": ('0.91r', '0.96r'), "ORF_reverse" : ('0.83r', '0.88r'), "GC" :  ('0.68r', '0.8r'), "links" : '0.64r'},
                                {"karyotype": '0.80r', "ORF": ('0.91r', '0.96r'), "ORF_reverse" : ('0.83r', '0.88r'), "custom_histogram" : ('0.68r', '0.8r'), "links" : '0.64r'},
                                {"karyotype": '0.80r', "GC" : ('0.84r', '0.96r'), "custom_histogram" : ('0.68r', '0.80r'), "links" : '0.64r'},
                                {"karyotype": '0.80r', "ORF": ('0.91r', '0.96r'), "ORF_reverse" : ('0.83r', '0.88r'), "GC" : ('0.68r', '0.8r'), "custom_histogram" : ('0.52r', '0.64r'), "links" : '0.48r'}]

    if (orfs == False and gc == False and custom_histogram == False):
        return (0, layer_radii_combinations[0])

    elif (orfs == True and gc == False and custom_histogram == False):
        return (1, layer_radii_combinations[1])

    elif (orfs == False and gc == True and custom_histogram == False):
        return (2, layer_radii_combinations[2])

    elif (orfs == False and gc == False and custom_histogram == True):
        return (3, layer_radii_combinations[3])

    elif (orfs == True and gc == True and custom_histogram == False):
        return (4, layer_radii_combinations[4])

    elif (orfs == True and gc == False and custom_histogram == True):
        return (5, layer_radii_combinations[5])

    elif (orfs == False and gc == True and custom_histogram == True):
        return (6, layer_radii_combinations[6])

    elif (orfs == True and gc == True and custom_histogram == True):
        return (7, layer_radii_combinations[7])



def create_circos_conf_file(output, karyotype_file, links_file, orfs, gc, custom_histogram, orf_forward_file=None, orf_reverse_file=None, gc_file=None, custom_histogram_file=None):
    """
    Creates the conf file for a circos diagram based on the type and amount of layers.

    Args:
        output (str): Output file path for the circos diagram (e.g. path/circos.conf)
        karyotype_file (str): File path to karyotype.txt file
        orfs (bool): An orf layer is inlcuded in the diagram (True or False)
        gc (bool): A GC layer is inlcuded in the diagram (True or False)
        custom_histogram (bool): A custom histogram layer is inlcuded in the diagram (True or False)
        orf_forward_file (str): File path to ORF.txt file
        orf_reverse_file (str): File path to ORF_reverse.txt file
        gc_file (str): File path to gc_content.txt if GC track supposed to be inlcuded\
        custom_histogram_file (str): File path to txt file data for the custom histogram layer
        links_file (str): File path to links.txt file

    Returns:
        None

    Notes:
        Layer config variable holds the layer combination, which are outlined below:
            0 = karyotype, links
            1 = karyotype, ORFs, links
            2 = karyotype, GC, links
            3 = karyotype, custom_histogram, links
            4 = karyotype, ORFs, GC, links
            5 = karyotype, ORFs, custom_histogram, links
            6 = karyotype, GC, custom_histogram, links
            7 = karyotype, ORFs, GC, custom_histogram, links
    """

    # Determine the radii for the given layer configuration
    layer_radii_and_type = find_layer_radii(orfs, gc, custom_histogram)
    layer_config = layer_radii_and_type[0]
    layer_radii_dict = layer_radii_and_type[1]

    # Erase everything in the conf file
    try:
        open(output, 'w').close()
    except Exception as e:
        print(e)
        print("Was not able to create the circos.conf configuration file")

    # Open the file in append mode
    try:
        f = open(output, "a")
    except Exception as e:
        print(e)
        print("Was not able to create the circos.conf configuration file.")


    f.write(
"""
################################################################
# Ideogram and Karyotype Setup
################################################################

karyotype = """ + karyotype_file + """

<ideogram>

<spacing>
default = 0.01r
</spacing>

# Ideogram position, fill and outline
radius    = 0.80r
thickness = 90p
fill      = yes
stroke_color = black
stroke_thickness = 8p

show_bands = yes
fill_bands = yes
band_transparency = 0
show_label = yes
# see etc/fonts.conf for list of font names
label_font = bold
label_radius = 1r + 130p
label_size = 55
label_parallel = yes
</ideogram>


################################################################
# Ticks and labels on fosmid karyotype bars
################################################################

show_ticks = yes
show_tick_labels = yes

<ticks>
skip_first_label = no
skip_last_label = no
radius = dims(ideogram,radius_outer)
# Setting multiplier to this means 10 000 will show as 10
multiplier = 1e-3
color = black
thickness = 4p
size = 20p

# 5kb bars with no label
<tick>
spacing = 5000u
size = 0.2r
show_label = no
label_size = 1r
thickness = 4p
color = black
</tick>

# 10kb bars with labels
<tick>
spacing = 10000u
size = 0.3r
show_label = yes
label_size = 1r
label_offset = 0.5r
thickness = 6p
color = black
</tick>
</ticks>

################################################################
# Beginning of Plots Section
################################################################
<plots>
"""
    )

    if layer_config in (1,4,5,7):
        f.write(
"""
################################################################
# ORF Forward and Reverse Layers
################################################################

# Forward Strand ORFs
<plot>
type            = tile
layers_overflow = hide
file        = """ + orf_forward_file + """
r1          = """ + layer_radii_dict['ORF'][1] + """
r0          = """ + layer_radii_dict['ORF'][0] + """
orientation = in
layers      = 3
margin      = 0.02u
thickness   = 15
padding     = 8
stroke_thickness = 1
stroke_color   = black
color = outer_orf_blue_color
<backgrounds>
<background>
color = outer_orf_background_color
</background>
</backgrounds>
</plot>

# Reverse Strand ORFs
<plot>
type            = tile
layers_overflow = hide
file        = """ + orf_reverse_file + """
r1          = """ + layer_radii_dict['ORF_reverse'][1] + """
r0          = """ + layer_radii_dict['ORF_reverse'][0] + """
orientation = in
layers      = 3
margin      = 0.02u
thickness   = 15
padding     = 8
stroke_thickness = 1
stroke_color     = black
color = inner_orf_red_color
<backgrounds>
<background>
color = inner_orf_background_color
</background>
</backgrounds>
</plot>
"""
        )
    if layer_config in (2,4,6,7):
        f.write(
"""
################################################################
# GC Content Layer
################################################################

<plot>
#Histogram Data Format: chr start end value [options]
type      = histogram
file      = """ + gc_file + """
r1        = """ + layer_radii_dict['GC'][1] + """
r0        = """ + layer_radii_dict['GC'][0] + """
max       = 100
min       = 0
stroke_type = outline
thickness   = 1
color       = vdgrey
extend_bin  = no
<backgrounds>
<background>
color = gc_histogram_background_color
</background>
</backgrounds>
<axes>
<axis>
spacing   = 0.1r
color     = vlgrey
thickness = 1
</axis>
</axes>
fill_color = gc_histogram_color
</plot>
"""
        )

    if layer_config in (3,5,6,7):
        f.write(
"""
################################################################
# Custom Histogram Layer
################################################################

<plot>
#Histogram Data Format: chr start end value [options]
type      = histogram
file      = """ + custom_histogram_file + """
r1        = """ + layer_radii_dict['custom_histogram'][1] + """
r0        = """ + layer_radii_dict['custom_histogram'][0] + """
max       = 100
min       = 0
stroke_type = outline
thickness   = 1
color       = vdgrey
extend_bin  = no
<backgrounds>
<background>
color = gc_histogram_background_color
</background>
</backgrounds>
<axes>
<axis>
spacing   = 0.1r
color     = vlgrey
thickness = 1
</axis>
</axes>
fill_color = gc_histogram_color
</plot>

"""
        )

    f.write(
"""
################################################################
# End of Plots Section
################################################################
</plots>

################################################################
# Links
################################################################
<links>
<link>
file = """ + links_file + """
radius = """ + layer_radii_dict['links'] + """
bezier_radius = 0.05r
thickness = 1
ribbon = yes
stroke_color = black_a4
stroke_thickness = 4
</link>
</links>

################################################################
# The remaining content is standard and required. It is imported
# from default files in the Circos distribution.
################################################################

<image>
<<include etc/image.conf>> # Included from Circos distribution.
</image>

# RGB/HSV color definitions, color lists, location of fonts, fill patterns.
# Included from Circos distribution.
<<include etc/colors_fonts_patterns.conf>>

# Debugging, I/O and other system parameters
# Included from Circos distribution.
<<include etc/housekeeping.conf>>

################################################################
# Defining custom colors
################################################################
<colors>
outer_orf_background_color = 120,120,120,0.30
inner_orf_background_color = 120,120,120,0.2
gc_histogram_background_color = 120,120,120,0.1
outer_orf_blue_color = vdgrey
inner_orf_red_color = vdgrey
gc_histogram_color = vdgrey
</colors>
"""
    )
    f.close()


def prtn_domain_legend(karyotype, file_name, ncol):
    """
    Take a karyotype.txt file and produce a legend for all the unique bands

    Args:
        karyotype (str): File path to a karyotype.txt file
        file_name (str): File name to be saved (should be name.png)
        ncol (int): Number of coloumns of labels

    Returns:
        None
    """
    with open(karyotype, 'r') as f:

        domains = {}

        for line in f:
            if line.startswith("band"):
                domains[line.split()[2]] = line.split()[6].replace("rgb", "")

        domains_list = list(domains.keys())
        rgb = [tuple(map(int, color[1:-1].split(','))) for color in list(domains.values())]

        draw_prtn_domain_legend(domains_list, rgb, file_name, ncol)


def draw_prtn_domain_legend(labels, colors, file_name, ncol):
    """
    Takes in labels and corresponding rgb colors and produces a legend

    Args:
        labels (list<str>): A list of labels
        Colors (list<tuples>): A list of rbg tubles (e.g. [(255,10,11), (56,110,255)]
        file_name (str): File name to be saved (should be name.png)
        ncol (int): Number of coloumns of labels

    Returns:
        None
    """

    # Convert colors to 0-1 scale
    normalized_colors = []
    for color in colors:
        normalized_colors.append((color[0]/255, color[1]/255, color[2]/255))

    plt.figure(figsize=(12,6))
    f = lambda m,c: plt.plot([],[],marker=m, color=c, ls="none")[0]
    plt.axis('off')
    handles = [f("s", normalized_colors[i]) for i in range(len(labels))]
    legend = plt.legend(handles, labels, loc=10, framealpha=1, frameon=False, ncol=ncol, mode="None", prop={'size': 6})

    fig  = legend.figure
    fig.canvas.draw()
    bbox  = legend.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(file_name, dpi=400, bbox_inches=bbox)


def karyotype_legend(karyotype_file, output_file):
    """
    Takes in a list of fosmid names with their corresponding number (e.g. 1 : FosmidA7)
    and produces a legend.

    Args:
        karyotype_file (str): File path to karyotype.txt file containing fosmid information
        output_file (str): File path to file where the legend is to be saved to (e.g. <path>/name.png)

    Returns:
        None
    """

    fosmid_names = []

    with open(karyotype_file, 'r') as f:

        for line in f:
            if line.startswith("chr"):
                fosmid_names.append(line.split()[3] + ' : ' + line.split()[2])

    plt.figure(figsize=(12,6))
    f = lambda : plt.plot([],[],marker=None, color='white', ls="none")[0]
    plt.axis('off')
    handles = [f() for i in range(len(fosmid_names))]
    legend = plt.legend(handles, fosmid_names, loc=9, framealpha=1, frameon=False, mode="None", prop={'size': 6}, handletextpad=0, borderpad=1)

    fig  = legend.figure
    fig.canvas.draw()
    bbox  = legend.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(output_file, dpi=400, bbox_inches=bbox)



"""All functions related to ORF data"""

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.SeqUtils import GC
from io import StringIO
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import pandas as pd
import itertools
import subprocess
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import statistics
import seaborn as sns



def create_orf_data(project_directory, correct_size_contigs_path):
    """
    Calls the functions to run prodigal, parse the output and write the results to
    txt files.

    Args:
        project_directory (str): A file path to the project_directory
        correct_size_contigs_path (str): A file path to the file containing all the correct sized contigs together

    Returns:
        prodigal_prtn_seq_output (str): The file path to prodigal's prtn translation output (.faa file)
    """

    print("Creating ORF data...")

    prodigal_output = project_directory + '/intermediate_outputs/prodigal/prodigal_orf_output'
    prodigal_prtn_seq_output = project_directory + '/intermediate_outputs/prodigal/prodigal_prtn_seq_output.faa'
    run_prodigal(correct_size_contigs_path, prodigal_output, prodigal_prtn_seq_output)

    prodigal_forward_and_reverse_df = parse_prodigal_sco_output(prodigal_output)

    prodigal_forward_and_reverse_df[0].to_csv(project_directory + '/circos_diagram_input_data/ORF.txt', sep=' ', index=False, header=False)
    prodigal_forward_and_reverse_df[1].to_csv(project_directory + '/circos_diagram_input_data/ORF_reverse.txt', sep=' ', index=False, header=False)

    return prodigal_prtn_seq_output


def run_prodigal(input, output, prtn_output, format='sco'):
    """
    Runs prodigal on the input .fasta file

    Args:
        input (str): File path to an input file (.fasta) to run prodigal on
        output (str): File path to output file (in format of format paramter)
        prtn_output (str): File path to a output file contianing protein translations (should be a .faa file)
        format (str): output format of the output file (default = sco)
            gbk:  Genbank-like format
            gff:  GFF format
            sqn:  Sequin feature table format
            sco:  Simple coordinate output

    Returns:
        None

    Notes:
        -c flag (Closed ends.  Do not allow genes to run off edges.) was not used.
        - When the ORF is indicated as -1 (or reverse) this means the orf was found on the reverse complement,
          in this case the indicated start is actually where there gene would have ended (just going in the other direction) and the
          end is where it would have started relative to the input sequence
        - Prodigal uses GTG as M (start codon)
        - Sequence must be 20000 characters for prodigal to run
    """
    prodigal_stdout_log_path = os.path.dirname(output) + '/prodigal_stdout_and_stderr_log.txt'

    prodigal_command = "prodigal -i " + input + " -o " + output + " -a " + prtn_output + " -f " + format
    with open(prodigal_stdout_log_path,"wb") as stdout:
        subprocess.call([str(prodigal_command)], stdout=stdout, stderr=subprocess.STDOUT, shell=True)

def parse_prodigal_sco_output(prodigal_sco_output):
    """
    Parses prokka sco format output file into a pandas DataFrame

    Args:
        prodigal_sco_output (str): File path to a prokka sco format ouptut file

    Returns:
        Two pandas DataFrames (forward strand genes, reverse strand genes)
            ['contig', 'orf_start', 'orf_end']

    Notes:
    - For reverse_strand ORFs, the orf_start is really the end (because it is in reverse)
    """

    with open(prodigal_sco_output, 'r') as f:

        forward_strand_orf_df = pd.DataFrame(index=[], columns=['contig', 'orf_start', 'orf_end'])
        reverse_strand_orf_df = pd.DataFrame(index=[], columns=['contig', 'orf_start', 'orf_end'])
        curContig = ""

        for line in f:
            if (line.startswith("# Sequence Data")):
                curContig = line.split('seqhdr="', 1)[1].rstrip()[:-1].split()[0]

            if (line.startswith(">")):
                start = line.split("_")[1]
                end = line.split("_")[2]
                strand = line.split("_")[3].rstrip()

                if strand == '+':
                    # add 1 to orf_end to adjust for circos indexing
                    new_forward_orf = {'contig' : curContig, 'orf_start' : int(start), 'orf_end' : int(end) + 1}
                    forward_strand_orf_df = forward_strand_orf_df.append(new_forward_orf, ignore_index=True)

                if strand == '-':
                    # add 1 to orf_end (really the start of the reverse strand orf) to adjust for circos indexing
                    new_reverse_orf = {'contig' : curContig, 'orf_start' : int(start), 'orf_end' : int(end) + 1}
                    reverse_strand_orf_df = reverse_strand_orf_df.append(new_reverse_orf, ignore_index=True)

        # get all values as ints
        forward_strand_orf_df['orf_start'] = forward_strand_orf_df['orf_start'].astype(int)
        forward_strand_orf_df['orf_end'] = forward_strand_orf_df['orf_end'].astype(int)
        reverse_strand_orf_df['orf_start'] = reverse_strand_orf_df['orf_start'].astype(int)
        reverse_strand_orf_df['orf_end'] = reverse_strand_orf_df['orf_end'].astype(int)

        return forward_strand_orf_df, reverse_strand_orf_df



import pathlib
from setuptools import find_packages, setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="fosvis",
    version="0.0.15",
    description="A tool for visualizing fosmid DNA sequences.",
    long_description=README,
    long_description_content_type="text/markdown",
    # url="https://github.com/realpython/reader",
    author="Tylo Roberts",
    author_email="tylojroberts@gmail.com",
    license="MIT",
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    install_requires=['biopython', 'pandas', 'numpy', 'matplotlib', 'seaborn', 'tqdm']
)



""""
All functions related to link data
"""

from Bio.Blast.Applications import NcbiblastnCommandline
from Bio.SeqUtils import GC
from io import StringIO
from Bio.Blast import NCBIXML
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import SeqIO
import pandas as pd
import itertools
import subprocess
import os
import sys
import numpy as np
import matplotlib.pyplot as plt
import statistics
import seaborn as sns
from tqdm import tqdm


def get_link_data(list_of_correct_len_contigs_paths, blast_type, project_directory, min_blast_similarity_percentage, min_blast_similarity_length, link_transperacny, percent_link_overlap_tolerance, keep_blast_data):

    """
    Returns a dataframe containing all the seq similarity link information.
    Colors links that overlap >= percent_link_overlap_tolerance with the same color

    Args:
        list_of_correct_len_contigs_paths list<str>: A list of correct length contig file names (full path, not just file name)
        blast_type (str): The type of blast to be used to create the links (i.e. 'blastn' OR 'tblastx')
        project_directory (str): File path to project directory
        min_blast_similarity_percentage (int): Only inlcude nucleotide similarity for link data with a percent identiity (similarity) >= to this
        min_blast_similarity_length (int): Only include nucleotide similarity for link data with an align_length (length of similarity) >= to this variable
        link_transperacny (str): The transperancy of the links in the diagram in range 0 - 1 (0 is transperant)
        percent_link_overlap_tolerance (int): The amount of overlap percentage required to classify two links as similar enough to color them with the same color
        keep_blast_data (bool): If True will keep the raw blast data, won't if false

    Returns:
        pandas.dataframe: A dataframe with nucleotide homology link data in the form:
            ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos', 'color_thickness']

    Notes:
        - The start position for the origin and terminus does not nessearily have to be bigger than the end pos
    """
    print("Creating Link data...")

    if (blast_type != 'blastn') and (blast_type != 'tblastx'):
        sys.exit("The blast type: " + blast_type  + "is not 'blastn' or 'tblastx'.  It needs to be one of these two.")

    all_homology_links_df = pd.DataFrame(index=[], columns=[])

    # progress bar
    total_len = sum(1 for e in itertools.combinations(list_of_correct_len_contigs_paths, 2))
    with tqdm(total=total_len) as pbar:
        # Loop through all combinations of contigs
        # for pair in tqdm(itertools.combinations(list_of_correct_len_contigs_paths, 2)):
        for pair in itertools.combinations(list_of_correct_len_contigs_paths, 2):
            if blast_type == 'blastn':
                pair_homology_df = get_pair_blastn_links(pair[0], pair[1], project_directory, min_blast_similarity_percentage, min_blast_similarity_length, keep_blast_data)
            elif blast_type == 'tblastx':
                pair_homology_df = get_pair_tblastx_links(pair[0], pair[1], project_directory, min_blast_similarity_percentage, min_blast_similarity_length, keep_blast_data)

            # Add new links to main df
            df_to_concat = [all_homology_links_df, pair_homology_df]
            all_homology_links_df = pd.concat(df_to_concat)

            # progress bar
            pbar.update(1)

    # Reset Indexes
    all_homology_links_df = all_homology_links_df.reset_index(drop=True)

    # adjust for circos_indexing
    all_homology_links_df['origin_end_pos'] += 1
    all_homology_links_df['terminus_end_pos'] += 1

    return color_similar_links(all_homology_links_df, link_transperacny, percent_link_overlap_tolerance)


def get_pair_blastn_links(query, subject, project_directory, min_blast_similarity_percentage, min_blast_similarity_length, keep_blast_data):
    """
    Indentify regions of nucleotide homology between two SeqRecord objects (greater
    than 'min_blast_similarity_percentage' across intervals of more than
    'min_blast_similarity_length' bp) and return a dataframe containing the infomration.

    Args:
        query (str): File path to a contig fasta file to be the query
        subject (str): File path to a contig fasta file to be the subject
        project_directory (str): File path for the project directory
        min_blast_similarity_percentage (int): Only use results with a percent identiity >= to this variable
        min_blast_similarity_length (int): Only use results that have an align_length >= to this variable
        keep_blast_data (bool): If True will keep the raw blast data, won't if false

    Returns:
        A dataframe conatining the homolgy information with the following structure is returned with each row repersenting a differnt alignment:
            ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos']

    Notes:
    - In the output the returned origin/terminus start positions do not have to be > than the end pos
    """

    # Run BLAST
    blastn_out = project_directory + "/intermediate_outputs/blastn/blastn_" + os.path.basename(query).replace('.fasta', '') + "_to_" + os.path.basename(subject).replace('.fasta', '') + "_output.xml"
    output = NcbiblastnCommandline(query=query, subject=subject, outfmt=5, out=blastn_out)()[0]

    # Read blastn output
    with open(blastn_out, "r") as blast_results_file:
        blast_result_record = NCBIXML.read(blast_results_file)

    seq_homology_df = pd.DataFrame(index=[], columns=['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos'])

    # Parse results
    for alignment in blast_result_record.alignments:
        subject_id = alignment.hit_def.split()[0]
        query_id = blast_result_record.query.split()[0]

        for hsp in alignment.hsps:
            identities_percentage = (hsp.identities / hsp.align_length) * 100

            # Append to df if similar/long enough
            if (identities_percentage >= min_blast_similarity_percentage) and (hsp.align_length >= min_blast_similarity_length):
                new_alignment = {'origin_var' : query_id, 'origin_start_pos' : hsp.query_start, 'origin_end_pos' : hsp.query_end, 'terminus_var' : subject_id, 'terminus_start_pos' : hsp.sbjct_start, 'terminus_end_pos' : hsp.sbjct_end}
                seq_homology_df = seq_homology_df.append(new_alignment, ignore_index=True)

    if keep_blast_data == False:
        os.remove(blastn_out)

    return seq_homology_df


def get_pair_tblastx_links(query, subject, project_directory, min_blast_similarity_percentage, min_blast_similarity_length, keep_blast_data):
    """
    Indentify regions of protein similarity using tblastx between two SeqRecord objects (greater
    than 'min_blast_similarity_percentage' similarity across intervals of more than
    'min_blast_similarity_length' bp) and return a dataframe containing the infomration.

    Args:
        query (str): File path to a contig fasta file to be the query
        subject (str): File path to a contig fasta file to be the subject
        project_directory (str): File path for the project directory
        min_blast_similarity_percentage (int): Only use results with a percent identiity >= to this variable
        min_blast_similarity_length (int): Only use results that have an aligned length >= to this variable
        keep_blast_data (bool): If True will keep the raw blast data, won't if false

    Returns:
        A dataframe conatining the homolgy information with the following structure is returned with each row repersenting a differnt alignment:
            ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos']

    Notes:
    - CHECK THIS In the output the returned origin/terminus start positions do not have to be > than the end pos
    - Alignment length is calculated abs(Query_End - Query_Start)
    """

    # Run Create output
    tblastx_out = project_directory + "/intermediate_outputs/tblastx/tblastx_" + os.path.basename(query) + "_to_" + os.path.basename(subject) + "_output.txt"

    # Run tblastx
    tblastx_command = 'bl2seq -p tblastx -i ' + query + ' -j ' + subject + ' -o ' + tblastx_out
    subprocess.call([str(tblastx_command)], shell=True)

    # Parse output for all matches
    results = parse_tblastx(tblastx_out)

    similarity_filt = results['Identities_Percent'] >= min_blast_similarity_percentage
    length_filt = abs(results['Query_End'] - results['Query_Start']) >= min_blast_similarity_length

    results = results[similarity_filt & length_filt]

    # Remove unnesary columns from results
    results.drop('Score', axis=1, inplace=True)
    results.drop('Expect', axis=1, inplace=True)
    results.drop('Identities_Percent', axis=1, inplace=True)
    results.drop('Positives_Percent', axis=1, inplace=True)

    # Arrange columns in correct order
    results = results[['Query', 'Query_Start', 'Query_End', 'Subject', 'Sbjct_Start', 'Sbjct_End']]

    # Change name of the columns
    results.columns = ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos']

    if keep_blast_data == False:
        os.remove(tblastx_out)

    return results


def parse_tblastx(input):
    """
    Parse a tblastx output into a pandas df output.
    Meant for an ouptut from 2 seqeunces

    Args:
        input (str):
        output (str):

    Returns:
        pandas df: Returns a pandas df with the columns [Query, Subject, Score, Expect, Identities_Percent, Positives_Percent,
                                                         Query_Start, Query_End, Sbjct_Start, Sbjct_End]

    Notes:
    - Score is in bits
    - Some e values don't have a number at beggining (e.g. e-163), a 1 is added to front (e.g. 1e-163)
    """

    with open(input, 'r') as f:
        content = f.readlines()

    # Remove white space
    content = [x.strip() for x in content]

    df_data = {
      'Query': [],
      'Subject': [],
      'Score': [],
      'Expect': [],
      'Identities_Percent': [],
      'Positives_Percent': [],
      'Query_Start': [],
      'Query_End': [],
      'Sbjct_Start': [],
      'Sbjct_End': []
    }

    # Dataframe Row
    Query = str(content[0].replace('Query= ', ''))
    Subject = 0
    Score = 0
    Expect = 0
    Identities_Percent = 0
    Positives_Percent = 0
    Query_Start = 0
    Query_End = 0
    Sbjct_Start = 0
    Sbjct_End = 0

    for line in content[1:]:

        # Get Sbjct seq name
        if line.startswith('>'):
            Subject = str(line.replace('>', ''))

        if line.startswith('Score'):
            # Write row to dataframe - signals at end of a data entry
            # Unless its the first run through
            if Score != '':
                df_data["Query"].append(Query)
                df_data["Subject"].append(Subject)
                df_data["Score"].append(Score)
                df_data["Expect"].append(Expect)
                df_data["Identities_Percent"].append(Identities_Percent)
                df_data["Positives_Percent"].append(Positives_Percent)
                df_data["Query_Start"].append(Query_Start)
                df_data["Query_End"].append(Query_End)
                df_data["Sbjct_Start"].append(Sbjct_Start)
                df_data["Sbjct_End"].append(Sbjct_End)

            # Get Score/Expect Data
            Score = float(line.replace('Score = ', '').split()[0])
            temp_expect = line.split(', ')[1].split()[2]
            if temp_expect.startswith('e'):
                Expect = float('1' + temp_expect)
            else:
                Expect = float(temp_expect)

        if line.startswith('Identities'):
            Identities_Percent = float(line.split('(')[1].split('%')[0])
            Positives_Percent = float(line.split(',')[1].split('(')[1].split('%')[0])

        if line.startswith('Query'):
            # If Query/Sbjct are multiple lines value will just be updated until
            # they are written to the df
            Query_Start = int(line.split()[1])
            Query_End = int(line.split()[3])

        if line.startswith('Sbjct'):
            # If Query/Sbjct are multiple lines value will just be updated until
            # they are written to the df
            Sbjct_Start = int(line.split()[1])
            Sbjct_End = int(line.split()[3])

    df = pd.DataFrame(df_data)

    return df


def color_similar_links(homolgy_links_df, link_transperacny, percent_link_overlap_tolerance):
    """
    Given a set of homology links, finds links that repersent the same seqeunce (they overlap with each other) being repersented by
    differnt links and color those with a matching unique color.

    Args:
        homolgy_links_df (pandas.df): A pandas dataframe repersenting all the nucleotide homology link data for a circos diagram
            ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos']
        link_transperacny (str): The transperancy of the links in the diagram in range 0 - 1 (0 is transperant)
        percent_link_overlap_tolerance (int): The amount of overlap percentage required to classify two links as similar enough to color them with the same color

    Returns:
        (pandas.df): A dataframe with link data colored accoridng to links that repersent the same sequence.
            ['origin_var', 'origin_start_pos', 'origin_end_pos', 'terminus_var', 'terminus_start_pos', 'terminus_end_pos', 'color_thickness']

    Notes:
        - Required that the dataframe indexes are from 0 - n in increasing order (groups are made based on indexes so they must all be distinct)
    """

    # List of link pairs that overlap sufficiently
    matching_link_indexes = []

    # Go thorugh all combinations of links - if overlap sufficiently, append to matching_link_indexes
    for index, link in homolgy_links_df.iterrows():

        for pot_match_index, pot_match_link in homolgy_links_df.iterrows():
            if pot_match_index != index: # don't compare to itself

                origin_percent_overlap = calc_percentage_overlap((link['origin_start_pos'], link['origin_end_pos']), (pot_match_link['origin_start_pos'], pot_match_link['origin_end_pos']))
                terminus_percent_overlap = calc_percentage_overlap((link['origin_start_pos'], link['origin_end_pos']), (pot_match_link['terminus_start_pos'], pot_match_link['terminus_end_pos']))

                # Case: origin of link to origin of pot_match_link
                if (link['origin_var'] == pot_match_link['origin_var']) and (origin_percent_overlap >= percent_link_overlap_tolerance): # same contig and overlap
                    matching_link_indexes.append([index, pot_match_index])

                # Case: origin of link to terminus of pot_match_link
                if (link['origin_var'] == pot_match_link['terminus_var']) and (terminus_percent_overlap >= percent_link_overlap_tolerance): # same contig and overlap
                    matching_link_indexes.append([index, pot_match_index])

        for pot_match_index, pot_match_link in homolgy_links_df.iterrows():
            if pot_match_index != index:

                origin_percent_overlap = calc_percentage_overlap((link['terminus_start_pos'], link['terminus_end_pos']), (pot_match_link['origin_start_pos'], pot_match_link['origin_end_pos']))
                terminus_percent_overlap = calc_percentage_overlap((link['terminus_start_pos'], link['terminus_end_pos']), (pot_match_link['terminus_start_pos'], pot_match_link['terminus_end_pos']))

                # Case: terminus of link to origin of pot_match_link
                if (link['terminus_var'] == pot_match_link['origin_var']) and (origin_percent_overlap >= percent_link_overlap_tolerance):
                    matching_link_indexes.append([index, pot_match_index])

                # Case: terminus of link to terminus of pot_match_link
                if (link['terminus_var'] == pot_match_link['terminus_var']) and (terminus_percent_overlap >= percent_link_overlap_tolerance):
                    matching_link_indexes.append([index, pot_match_index])


    # Put links that repersent the same sequence in sets
    link_indexs_w_common_homology = merge_lists_with_common_elements(matching_link_indexes)

    # Loop through all indexes in homolgy_links_df, if they are not in link_indexs_w_common_homology
    # then append a set with just that index i.e. {index} to another list and then combine with link_indexs_w_common_homology
    not_in_link_indexs_w_common_homology = []
    for index in list(homolgy_links_df.index.values):

        # Check if index is not in any of the sets in link_indexs_w_common_homology
        if not any(index in sublist for sublist in link_indexs_w_common_homology):
            not_in_link_indexs_w_common_homology.append({index})

    # Contains every index in a set by itself or a set with others
    link_indexs_w_common_homology = link_indexs_w_common_homology + not_in_link_indexs_w_common_homology

    # create seaborn color palette with len of link_indexs_w_common_homology
    color_palette = create_color_palette(len(link_indexs_w_common_homology))

    # Create dict with indexes as keys and colors as values - sets every index to same color as its group (if its in a group)
    link_index_color_dict = {}
    color_palette_counter = 0
    for index_set in link_indexs_w_common_homology:
        color = '(' + color_palette[color_palette_counter] + "," + link_transperacny + ')'
        for link_index in index_set:
            link_index_color_dict[link_index] = color
        color_palette_counter += 1

    # Go thorugh dict and set links to their correct color in homolgy_links_df
    for key, value in link_index_color_dict.items():
        homolgy_links_df.at[key, 'color_thickness'] = 'color=' + value + ',thickness=10p'

    return homolgy_links_df


def merge_lists_with_common_elements(listoflist):
    """
    Given a list of lists, merge all lists that share a common element, and repeat until there are no more lists with the same item.

    Args:
        listoflist (list<list<item>>): A list of lists to be merged

    Returns:
        (list<set>): A list of sets of merged items

    Source:
        Based off of https://stackoverflow.com/questions/4842613/merge-lists-that-share-common-elements
    """


    out = []
    while len(listoflist)>0:
        first, *rest = listoflist
        first = set(first)

        lf = -1
        while len(first)>lf:
            lf = len(first)

            rest2 = []
            for r in rest:
                if len(first.intersection(set(r)))>0:
                    first |= set(r)
                else:
                    rest2.append(r)
            rest = rest2

        out.append(first)
        listoflist = rest

    return out


def calc_percentage_overlap(range1, range2):
    """
    Calculates the percentage of overlap betwen two ranges.

    Args:
        range1 (tuple<int>): A range repersented as a tuple
        range2 (tuple<int>): A second range repersented as a tuple

    Returns:
        (int): The percentage overlap between ranges A and B

    Notes:
    - Ranges don't have to be passed in a smalles to largest value
    - The percentage of overlap between the two ranges is calculated by the
      following (simplified) formula:
        percentage_overlap = (width of overlaping region / total range of both ranges combined) * 100
    """

    A = range1
    B = range2

    if range1[0] > range1[1]:
        A = (range1[1], range1[0])

    if range2[0] > range2[1]:
        B = (range2[1], range2[0])

    return abs((statistics.median([A[0],B[0],B[1]])-statistics.median([A[1],B[0],B[1]])) / (max(A[1],B[1])-min(A[0],B[0]))) * 100


def create_color_palette(num_colors):
    """
    Creates a color palette with a specified amount of rgb colors.
    Uses the seaborn hls color palette

    Args:
        num_colors (int): The number of colors to be returned

    Returns:
        list<str>: A list of rgb tuples in the form "<num1>,<num2>,<num3>"
    """
    f = lambda y : tuple(map(lambda x : x*255, y))
    color_palette = [f(color) for color in sns.color_palette("hls", num_colors)]

    str_color_palette = []
    for color in color_palette:
        str_color_palette.append(",".join(map(str,color)))

    return str_color_palette



from fosvis import protein_domains

import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import filecmp
import os
import pandas as pd
from pandas.testing import assert_frame_equal
import difflib


class test_run_and_parse_hmmscan(unittest.TestCase):

    def test_run_and_parse_hmmscan(self):
        hmm_db = '/mnt/nfs/sharknado/Sandbox/Tylo/databases/hmmer/hmmscan/Pfam-A.hmm'

        # Run run_hmmscan() on sample orfs
        sample_orfs = './test/data_for_testing/testing_protein_domains/sample_orfs_and_hmmscan_df/prodigal_prtn_seq_output.faa'
        output_dir = './test/data_for_testing/testing_protein_domains/temp_hmmscan_output'
        protein_domains.run_hmmscan(sample_orfs, hmm_db, output_dir, 0.01, 'hmmscan_testing')

        # check output file exits and is not empty
        tblout_path = output_dir + '/hmmscan_testing_tblout.txt'
        domtblout_path = output_dir + '/hmmscan_testing_domtblout.txt'
        out_file_path = output_dir + '/hmmscan_testing_out.txt'
        self.assertTrue(os.path.exists(tblout_path) and os.path.getsize(tblout_path) > 0)
        self.assertTrue(os.path.exists(domtblout_path) and os.path.getsize(domtblout_path) > 0)
        self.assertTrue(os.path.exists(out_file_path) and os.path.getsize(out_file_path) > 0)

        # Parse the output files using parse_hmmscan_domtblout()
        actual_output_df = protein_domains.parse_hmmscan_domtblout(domtblout_path)

        # Check df
        self.assertEqual(5, actual_output_df.shape[1])
        self.assertEqual(['target', 'orf_id', 'orf_start', 'orf_stop', 'target_description'], list(actual_output_df.columns.values))

        # Delete all files
        os.remove(tblout_path)
        os.remove(domtblout_path)
        os.remove(out_file_path)


class test_convert_hmmscan_domains_from_protein_to_nuc_pos(unittest.TestCase):

    def test_convert_hmmscan_domains_from_protein_to_nuc_pos(self):
        sample_hmmscan_df_path = './test/data_for_testing/testing_protein_domains/sample_orfs_and_hmmscan_df/sample_hmmscan_df.csv'
        sample_hmmscan_df = pd.read_csv(sample_hmmscan_df_path)
        sample_orfs = './test/data_for_testing/testing_protein_domains/sample_orfs_and_hmmscan_df/prodigal_prtn_seq_output.faa'
        converted_df_output = protein_domains.convert_hmmscan_domains_from_protein_to_nuc_pos(sample_hmmscan_df, sample_orfs)

        # Check df
        self.assertEqual(7, converted_df_output.shape[1])
        self.assertEqual(['band', 'contig', 'target_var', 'target_label', 'start', 'stop', 'color'], list(converted_df_output.columns.values))


class test_get_prtn_domain_band_data(unittest.TestCase):

    def test_get_prtn_domain_band_data(self):
        sample_orfs = './test/data_for_testing/testing_protein_domains/sample_orfs_and_hmmscan_df/prodigal_prtn_seq_output.faa'
        hmm_db = '/mnt/nfs/sharknado/Sandbox/Tylo/databases/hmmer/hmmscan/Pfam-A.hmm'
        output_dir = './test/data_for_testing/testing_protein_domains/temp_hmmscan_output'


        actual_df = protein_domains.get_prtn_domain_band_data(sample_orfs, hmm_db, output_dir, 0.01, 'testing_get_prtn_domain_band_data')

        self.assertEqual(7, actual_df.shape[1])
        self.assertEqual(['band', 'contig', 'target_var', 'target_label', 'start', 'stop', 'color'], list(actual_df.columns.values))


        # remove Files
        tblout_path = output_dir + '/testing_get_prtn_domain_band_data_tblout.txt'
        domtblout_path = output_dir + '/testing_get_prtn_domain_band_data_domtblout.txt'
        out_file_path = output_dir + '/testing_get_prtn_domain_band_data_out.txt'
        os.remove(tblout_path)
        os.remove(domtblout_path)
        os.remove(out_file_path)


class test_random_rgb_color(unittest.TestCase):

    def test_random_rgb_color(self):
        result = protein_domains.random_rgb_color()
        self.assertTrue(isinstance(result, str))
        self.assertEqual(result[:4], 'rgb(')


if __name__ == "__main__":
    unittest.main()
